import base64
import json
from collections import ChainMap
from typing import Any, Type

from botocore.exceptions import ClientError

from aws_chain_secrets.caster import cast
from aws_chain_secrets.client import Client
from aws_chain_secrets.exceptions import SecretsManagerException


class Null:
    """
    class to represent null type
    """
    pass


null = Null()


class SecretsManager(Client):
    def __init__(
            self,
            region_name: str,
            *secret_names: str,
            aws_access_key_id: str = None,
            aws_secret_access_key: str = None
    ):
        """
        initial setup
        :param region_name: AWS region name
        :param secret_names: These are the secret names from which to read settings. The last name has priority.
        """
        self.secret_names = secret_names
        super().__init__(
            "secretsmanager",
            region_name,
            aws_access_key_id,
            aws_secret_access_key
        )
        self._secrets = {secret_name: {} for secret_name in secret_names}
        self._combined = ChainMap(*reversed(self._secrets.values()))

    def __call__(self, key, *, default: Any = null, casting_type: Type = str):
        return self.get(key, default=default, casting_type=casting_type)

    def __enter__(self):
        super().__enter__()
        self.fetch()
        return self

    def get(self, key, *, default: Any = null, casting_type: Type = str):
        """
        get secret value
        :param key: key value to search
        :param default: default value if key not exists in secrets
        :param casting_type: casting type to convert value
        :return: Any secret value specified by key
        """
        try:
            value = self._combined[key]
        except KeyError:
            if isinstance(default, Null):
                raise KeyError(f"{repr(key)} not found. Please check secret value in AWS Secrets Manager.")
            return default
        return cast(value, casting_type)

    def set(self, secret_name: str, key, value):
        """
        set a `key: value` to the specified secret
        :param secret_name: secret_name of AWS Secrets Manager
        :param key: key of map
        :param value: value of map
        """
        self._secrets[secret_name][key] = value

    def keys(self):
        return self._combined.keys()

    def values(self):
        return self._combined.values()

    def items(self):
        return self._combined.items()

    def fetch(self):
        """
        get secrets from AWS and update
        """
        for secret_name in self.secret_names:
            self._secrets[secret_name].clear()
            self._secrets[secret_name].update(self._get_secrets(secret_name))

    def update(self, secret_name: str):
        """
        update specified secrets from local to AWS
        :param secret_name: secret name of AWS Secrets Manager
        """
        self._client.update_secret(SecretId=secret_name, SecretString=json.dumps(self._secrets[secret_name]))

    def _get_secrets(self, secret_name: str) -> dict:
        """
        get secrets from AWS
        :param secret_name: secret name of AWS Secrets Manager
        """
        try:
            response = self._client.get_secret_value(SecretId=secret_name)
        except ClientError as e:
            raise SecretsManagerException.from_code(e.response['Error']['Code'])
        else:
            if 'SecretString' in response:
                return json.loads(response['SecretString'])
            return json.loads(base64.b64decode(response['SecretBinary']))





from typing import Optional

import boto3
from botocore.client import BaseClient


class Client:
    """
    wrapper for AWS client
    """
    def __init__(
            self,
            service_name: str,
            region_name: str,
            aws_access_key_id: str = None,
            aws_secret_access_key: str = None
    ):
        self.service_name = service_name
        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self._client: Optional[BaseClient] = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def _create_client(self) -> BaseClient:
        return boto3.session.Session().client(
            service_name=self.service_name,
            region_name=self.region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key
        )

    def connect(self):
        self._client = self._create_client()

    def disconnect(self):
        self._client.close()
        del self._client





class SecretsManagerException(Exception):
    DEFAULT_MESSAGE = "Unknown error occurred."

    def __init__(self, message: str = None):
        super().__init__(message or self.DEFAULT_MESSAGE)

    @classmethod
    def from_code(cls, code: str, message: str = None) -> 'SecretsManagerException':
        if code == 'DecryptionFailureException':
            return DecryptionFailureException(message)
        elif code == 'InternalServiceErrorException':
            return InternalServiceErrorException(message)
        elif code == 'InvalidParameterException':
            return InvalidParameterException(message)
        elif code == 'InvalidRequestException':
            return InvalidRequestException(message)
        elif code == 'ResourceNotFoundException':
            return ResourceNotFoundException(message)
        else:
            return cls(message)


class DecryptionFailureException(SecretsManagerException):
    DEFAULT_MESSAGE = "Secrets Manager can't decrypt the protected secret text using the provided KMS key."


class InternalServiceErrorException(SecretsManagerException):
    DEFAULT_MESSAGE = "An error occurred on the server side."


class InvalidParameterException(SecretsManagerException):
    DEFAULT_MESSAGE = "You provided an invalid value for a parameter.",


class InvalidRequestException(SecretsManagerException):
    DEFAULT_MESSAGE = "You provided a parameter value that is not valid for the current state of the resource."


class ResourceNotFoundException(SecretsManagerException):
    DEFAULT_MESSAGE = "We can't find the resource that you asked for."



import unittest

from aws_chain_secrets import caster


class CasterTestCase(unittest.TestCase):
    def test_int(self):
        self.assertEqual(1, caster.cast('1', int))

    def test_bool(self):
        self.assertTrue(caster.cast('true', bool))

    def test_list(self):
        self.assertEqual(['1', '2', '3', '4'], caster.cast('1,2,3,4', list))
        self.assertEqual([1, 2, 3, 4], caster.cast('1, 2, 3, 4', list[int]))

    def test_set(self):
        self.assertEqual({1, 2, 3, 4}, caster.cast('1,2,3,4', set[int]))

    def test_dict(self):
        self.assertEqual({'foo': 'bar'}, caster.cast('{"foo": "bar"}', dict))



import json
import string
from dataclasses import dataclass
from types import GenericAlias
from typing import Any, Type, get_origin, get_args

TRUE_STRINGS = ('y', 'yes', 't', 'true', 'on', '1')
FALSE_STRINGS = ('', 'n', 'no', 'f', 'false', 'off', '0')


def string_to_boolean(value: str) -> bool:
    """
    Convert a string representation of truth to True or False.

    :param value: a string representation of truth
    :return: True if value represents truth, False if value represents false, otherwise raise ValueError
    :raises: Raises ValueError if 'value' is anything else
    """
    value = value.lower()
    if value in TRUE_STRINGS:
        return True
    if value in FALSE_STRINGS:
        return False
    raise ValueError(f"invalid truth value {repr(value)}")


def cast_boolean(value: Any) -> bool:
    value = str(value)
    return string_to_boolean(value)


def cast(value: str, casting_type: Type = None) -> Any:
    """
    string value to typed value
    :param value: original string
    :param casting_type: expected type to be
    :return: typed value
    """
    if casting_type is None:
        return value
    if casting_type is bool:
        return cast_boolean(value)
    if casting_type in (set, list, tuple, bytes, bytearray):
        return Sequence(wrapper=casting_type)(value)
    if casting_type is dict:
        return json.loads(value)
    if isinstance(casting_type, GenericAlias):
        origin = get_origin(casting_type)
        args = get_args(casting_type)
        if origin in (set, list, tuple, bytes, bytearray):
            return Sequence(cast_type=args[0] if args else str, wrapper=origin)(value)
        raise NotImplementedError(f"cannot cast to {repr(casting_type)}")
    return casting_type(value)


@dataclass
class Sequence:
    """
    Helper to cast Sequence (list, tuple, bytes, bytearray) and Set(set, frozenset)
    """
    cast_type: Type = str
    delimiter: str = ','
    strip: str = string.whitespace
    wrapper: Type = list

    def __call__(self, value: str):
        return self.wrapper(cast(v.strip(self.strip), self.cast_type) for v in value.split(self.delimiter))



import setuptools

with open("README.rst", "r", encoding='utf-8') as readme:
    long_description = readme.read()

setuptools.setup(
    name="aws_chain_secrets",
    version='1.0.0',
    author="Hyouk Oh",
    author_email="h.5.kure@gmail.com",
    description="This aims to enable the use of one or more secrets from AWS SecretsManager.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/h5kure/aws-chain-secrets",
    packages=setuptools.find_packages(),
    package_data={"": ["py.typed"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['boto3']
)
